import java.util.ArrayList;
import java.util.List;


/***
 * This class implements a rectangular map.
 * The map is divided into fixed sized squares called map points (MapPoint).
 * Each MapPoint has a unique coordinate.
 * The upper left map point has coordinate (0,0)
 * and the lower right map point (sizeX - 1, sizeY - 1)
 * where the width of the map is sizeY and the height is sizeX in number of squares.
 * (x,y)
 * (0,0)
 * -----------> Y
 * |
 * |
 * |
 * |
 * X
 *
 * Example 5x5
 * (0,0) (0,1) (0,2) (0,3) (0,4)
 * (1,0) (1,1) (1,2) (1,3) (1,4)
 * (2,0) (2,1) (2,2) (2,3) (2,4)
 * (3,0) (3,1) (3,2) (3,3) (3,4)
 * (4,0) (4,1) (4,2) (4,3) (4,4)
 *
 */
public class FactoryMap {
    private MapPoint[][] map;
    private int sizeX;
    private int sizeY;

    /**
     * Initializes a Map without any obstacles
     *
     * @param sizeX the height of the map
     * @param sizeY the width of the map
     */
    public FactoryMap(int sizeX, int sizeY) {
        this.sizeX = sizeX;
        this.sizeY = sizeY;
        createMap();
    }

    /**
     * Initializes a Map with the specified obstacles.
     *
     * @param sizeX the height of the map
     * @param sizeY the width of the map
     */
    public FactoryMap(int sizeX, int sizeY, Obstacle[] obstacles) {
        this(sizeX, sizeY);
        setObstacles(obstacles);
    }

    private void createMap() {
        map = new MapPoint[sizeX][sizeY];
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                map[xCor][yCor] = new MapPoint(xCor, yCor);
            }
        }
    }

    public void setObstacle(Obstacle obst) {
        for (int xCor = obst.x; xCor < obst.x + obst.dx; xCor++) {
            for (int yCor = obst.y; yCor < obst.y + obst.dy; yCor++) {
                map[xCor][yCor].setBlocked();
            }
        }
    }

    public void setObstacles(Obstacle[] obstacles) {
        for (Obstacle obst : obstacles)
            setObstacle(obst);
    }

    public void setSlowedField(int x, int y) {
        map[x][y].setSlowed();
    }

    /**
     * Returns the MapPoint with the given Coordinates
     *
     * @param x
     * @param y
     * @return
     */
    public MapPoint getPoint(int x, int y) {
        return map[x][y];
    }

    /**
     * Returns all neighbouring MapPoints x where x is not blocked.
     * Bsp (# means blocked):
     * o o #
     * o i o
     * o o o
     * Here all o's would be returned when i is the input to the Method
     *
     * @param point the input MapPoint
     * @return a list of neighbouring points
     */
    public List<MapPoint> getMovableNeighbours(MapPoint point) {
        ArrayList<MapPoint> tmp = new ArrayList<>();
        //TODO
        return tmp;
    }

    public int countNumUnexplored() {
        int count = 0;
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                if (map[xCor][yCor].state == MapPoint.State.UNEXPLORED)
                    count++;
            }
        }
        return count;
    }

    public int getSizeX() {
        return sizeX;
    }

    public int getSizeY() {
        return sizeY;
    }


    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int xCor = 0; xCor < sizeX; xCor++) {
            for (int yCor = 0; yCor < sizeY; yCor++) {
                MapPoint point = map[xCor][yCor];
                sb.append(point.getSymbol());
                sb.append(" ");
            }
            sb.append("\n");
        }
        return sb.toString();
    }

    /**
     * Encodes a rectangular area, which is blocked.
     */
    public static class Obstacle {
        int x;
        int y;
        int dx;
        int dy;

        public Obstacle(int x, int y, int dx, int dy) {
            this.x = x;
            this.y = y;
            this.dx = dx;
            this.dy = dy;
        }
    }

    /**
     * Represents a Point on the Map.
     */
    public static class MapPoint {

        enum State {
            BLOCKED,
            EXPLORED,
            UNEXPLORED,
            CURRENT,
            START,
            TARGET,
            PATH
        }

        int x;
        int y;
        private State state;
        private boolean slowed = false;

        private MapPoint(int x, int y) {
            this.x = x;
            this.y = y;
            this.state = State.UNEXPLORED;
        }

        public void setExplored() {
            if (this.state == State.BLOCKED)
                throw new RuntimeException("A blocked node cannot be explored");
            this.state = State.EXPLORED;
        }

        public void setUnexplored() {
            if (this.state == State.BLOCKED)
                throw new RuntimeException("A blocked node cannot be unexplored");
            this.state = State.UNEXPLORED;
        }

        public void setPath() {
            this.state = State.PATH;
        }

        public void setTarget() {
            this.state = State.TARGET;
        }

        public void setCurrent() {
            this.state = State.CURRENT;
        }

        public void setBlocked() {
            this.state = State.BLOCKED;
        }

        public void setStart() {
            this.state = State.START;
        }

        public void setSlowed() {
            slowed = true;
        }

        /**
         * Computes the geometric distance between this point (x1, y1) and another point (x2, y2):
         * distance = sqrt((x2-x1)^2 + (y2- y1)^2)
         *
         * @param other the other MapPoint
         * @return the geometric distance to the other point
         */
        public double getDistance(MapPoint other) {
            //TODO
            return -1.0;
        }

        /**
         * Computes the geometric distance between this point (x1, y1) and another point (x2, y2), and adds
         * a slowingCost constant if one of them has state SLOWED, or 2x slowingCost if both of them are in state SLOWED
         * @param other the other MapPoint
         * @param slowingCost the constant distance cost of a slowed MapPoint
         * @return the geometric distance to the other point including slowing Costs
         */
        public double getDistanceWithSlowingFactor(MapPoint other, double slowingCost) {
            //TODO
            return -1.0;
        }

        public int getX() {
            return x;
        }

        public int getY() {
            return y;
        }

        public State getState() {
            return this.state;
        }

        public boolean isBlocked() {
            return this.state == State.BLOCKED;
        }

        public boolean isSlowed() {
            return this.slowed;
        }

        public char getSymbol() {
            switch (this.state) {
                case BLOCKED:
                    return '#';
                case EXPLORED:
                    if (this.isSlowed())
                        return 'O';
                    else
                        return 'o';
                case UNEXPLORED:
                    if (this.isSlowed())
                        return '~';
                    else
                        return '?';
                case START:
                    return 'S';
                case TARGET:
                    return 'T';
                case PATH:
                    return 'X';
                case CURRENT:
                    return 'C';
                default:
                    return '-';

            }
        }

        @Override
        public String toString() {
            String out =  String.format("x:%d y:%d s:%s", this.x, this.y, this.state.toString());
            return this.isSlowed() ? out + "/SLOWED" : out;
        }

        @Override
        public boolean equals(Object obj) {
            if (!(obj instanceof MapPoint))
                return false;
            MapPoint point = (MapPoint) obj;
            return point.x == this.x && point.y == this.y;
        }

        @Override
        public int hashCode() {
            int tmp = y + ((x + 1) / 2);
            return x + (tmp * tmp);
        }

    }

}
